import numpy as np
import pandas as pd
import pingouin
from numpy.typing import ArrayLike
from scipy import stats

from critical_es_value import utils


def critical_for_correlation_test_from_values(
    r: float,
    n: int,
    confidence: float = 0.95,
    alternative: str = "two-sided",
    variant: str = "ttest",
) -> pd.DataFrame:
    """
    Calculate critical effect size values given pearson correlation coefficient, sample size and significance level.

    Args:
        r (float): Pearson correlation coefficient.
        n (int): Sample size.
        confidence (float): Confidence level between 0 and 1 (exclusive). Default is 0.95.
        alternative (str): The alternative hypothesis. Either "two-sided", "greater", or "
        variant (str): The statistical test variant. Either "ttest" or "ztest". Default is "ttest".

    Returns:
        pd.DataFrame: A DataFrame with the following columns:
            - `r`: Pearson correlation coefficient
            - `n`: Sample size
            - `dof`: Degrees of freedom
            - `r_critical`: Critical value for the correlation coefficient
            - `rz_critical`: Critical value for Fisher's z-transformed correlation coefficient (only for "ztest" variant)
            - `se_r`: Standard error of the correlation coefficient
            - `se_r_critical`: Standard error of the critical correlation coefficient
            - `se_rz_critical`: Standard error of the critical Fisher's z-transformed correlation coefficient (only for "ztest" variant)

    Raises:
        ValueError: If `variant` is not one of "ttest" or "ztest".

    Examples:
        >>> import pingouin as pg
        >>> import critical_es_value as cev
        >>> x = [1,2,3,4,5,6]
        >>> y = [2,2,2,3,3,3]
        >>> corr = pg.corr(x, y, method="pearson").iloc[0]
        >>> cev.critical_for_correlation_test(x, y)
                  n        r  dof  r_critical      se_r  se_r_critical
        critical  6  0.87831    4    0.811401  0.239046       0.292245
    """
    if variant not in ["ttest", "ztest"]:
        raise ValueError("variant must be one of 'ttest' or 'ztest'")

    alpha = utils.get_alpha(confidence, alternative)
    dof = n - 2

    if variant == "ttest":
        tc = np.abs(stats.t.ppf(alpha, dof))
        rc = np.sqrt(tc**2 / (tc**2 + dof))
    else:
        zc = np.abs(stats.norm.ppf(alpha))
        rc = np.tanh(zc / np.sqrt(n - 3))

    result = {
        "n": n,
        "r": r,
        "dof": dof,
        "r_critical": rc,
        "se_r": np.sqrt((1 - r**2) / dof),
        "se_r_critical": np.sqrt((1 - rc**2) / dof),
    }
    if variant == "ztest":
        result["rz_critical"] = np.atanh(rc)
        result["se_rz_critical"] = 1 / np.sqrt(n - 3)

    return pd.DataFrame([result], index=["critical"])


def critical_for_correlation_test(
    x: ArrayLike,
    y: ArrayLike,
    confidence: float = 0.95,
    alternative: str = "two-sided",
    variant: str = "ttest",
) -> pd.DataFrame:
    """
    Calculate critical effect size values for a pearson correlation test.

    Args:
        x (ArrayLike): Sample data for group 1.
        y (ArrayLike): Sample data for group 2.
        confidence (float): Confidence level between 0 and 1 (exclusive). Default is 0.95.
        alternative (str): The alternative hypothesis. Either "two-sided", "greater", or "less". Default is "two-sided".
        variant (str): The statistical test variant. Either "ttest" or "ztest". Default is "ttest".

    Returns:
        pd.DataFrame: A DataFrame with the following columns:
            - `r`: Pearson correlation coefficient
            - `n`: Sample size
            - `dof`: Degrees of freedom
            - `r_critical`: Critical value for the correlation coefficient
            - `rz_critical`: Critical value for Fisher's z-transformed correlation coefficient (only for "ztest" variant)
            - `se_r`: Standard error of the correlation coefficient
            - `se_r_critical`: Standard error of the critical correlation coefficient
            - `se_rz_critical`: Standard error of the critical Fisher's z-transformed correlation coefficient (only for "ztest" variant)

    Examples:
        >>> import critical_es_value as cev
        >>> x = [1,2,3,4,5,6]
        >>> y = [2,2,2,3,3,3]
        >>> cev.critical_for_correlation_test(x, y)
                  n        r  dof  r_critical      se_r  se_r_critical
        critical  6  0.87831    4    0.811401  0.239046       0.292245
    """
    corr = pingouin.corr(x, y, alternative=alternative, method="pearson").iloc[0]

    return critical_for_correlation_test_from_values(
        r=corr["r"],
        n=corr["n"],
        confidence=confidence,
        alternative=alternative,
        variant=variant,
    )



from typing import Optional

import numpy as np
import pandas as pd
import pingouin
from numpy.typing import ArrayLike
from scipy import stats

from critical_es_value import utils


def critical_for_linear_regression_from_values(
    coeffs: ArrayLike,
    coeffs_se: ArrayLike,
    coeffs_names: ArrayLike,
    dof: Optional[int] = None,
    confidence: float = 0.95,
    alternative: str = "two-sided",
    variant: str = "ttest",
) -> pd.DataFrame:
    """Calculate critical effect size values given linear regression coefficients.

    Args:
        coeffs (ArrayLike): Estimated regression coefficients.
        coeffs_se (ArrayLike): Standard errors of the regression coefficients.
        coeffs_names (ArrayLike): Names of the regression coefficients.
        dof (Optional[int]): Degrees of freedom of the model residuals. Only used for "ttest" variant.
        confidence (float): Confidence level between 0 and 1 (exclusive). Default is 0.95.
        alternative (str): The alternative hypothesis. Either "two-sided", "greater", or "
        variant (str): The statistical test variant. Either "ttest" or "ztest". Default is "ttest".

    Returns:
        pd.DataFrame: An array containing critical effect size values for each coefficient.

    Raises:
        ValueError: If variant is not one of "ttest" or "ztest".

    Examples:
        >>> import pandas as pd
        >>> import pingouin as pg
        >>> import critical_es_value as cev
        >>> df = pd.DataFrame({
        ...    "x1": [1,2,3,4,5,6],
        ...    "x2": [2,2,2,3,3,3],
        ...    "y": [3,4,5,6,7,8],
        ... })
        >>> model = pg.linear_regression(df[["x1", "x2"]], df["y"])
        >>> cev.critical_for_linear_regression_from_values(
        ...     coeffs=model["coef"].values,
        ...     coeffs_se=model["se"].values,
        ...     coeffs_names=model["names"].values,
        ...     dof=model.df_resid_,
        ...     variant="ttest",
        ... )
               names          coef  coef_critical
        0  Intercept  2.000000e+00   1.082596e-14
        1         x1  1.000000e+00   1.875111e-15
        2         x2  1.527841e-15   6.404723e-15
    """
    alpha = utils.get_alpha(confidence, alternative)

    if variant not in ["ttest", "ztest"]:
        raise ValueError("variant must be one of 'ttest' or 'ztest'")

    if variant == "ttest" and dof is None:
        raise ValueError("dof must be provided for 'ttest' variant")

    if variant == "ttest":
        qc = np.abs(stats.t.ppf(alpha, dof))
    else:
        qc = np.abs(stats.norm.ppf(alpha))

    return pd.DataFrame(
        {
            "names": coeffs_names,
            "coef": coeffs,
            "coef_critical": qc * np.array(coeffs_se),
        },
        index=list(range(len(coeffs))),
    )


def critical_for_linear_regression(
    X: pd.DataFrame,
    y: pd.Series,
    confidence: float = 0.95,
    alternative: str = "two-sided",
    variant: str = "ttest",
    **kwargs,
) -> pd.DataFrame:
    """Calculate critical effect size values for linear regression coefficients.

    Args:
        X (pd.DataFrame): DataFrame containing the independent variables.
        y (pd.Series): Series containing the dependent variable.
        confidence (float): Confidence level between 0 and 1 (exclusive). Default is 0.95.
        alternative (str): The alternative hypothesis. Either "two-sided", "greater", or "less". Default is "two-sided".
        variant (str): The statistical test variant. Either "ttest" or "ztest". Default is "ttest".
        **kwargs: Additional keyword arguments to pass to pingouin.linear_regression.

    Returns:
        pd.DataFrame: Returns a DataFrame with the following columns:
            - `names`: Names of the regression coefficients
            - `coef`: Estimated regression coefficients
            - `coef_critical`: Critical value for the regression coefficients

    Examples:
        >>> import pandas as pd
        >>> import critical_es_value as cev
        >>> df = pd.DataFrame({
        ...    "x1": [1,2,3,4,5,6],
        ...    "x2": [2,2,2,3,3,3],
        ...    "y": [3,4,5,6,7,8],
        ... })
        >>> cev.critical_for_linear_regression(df[["x1", "x2"]], df["y"])
               names          coef  coef_critical
        0  Intercept  2.000000e+00   1.082596e-14
        1         x1  1.000000e+00   1.875111e-15
        2         x2  1.527841e-15   6.404723e-15
    """
    alpha = utils.get_alpha(confidence, alternative)
    model = pingouin.linear_regression(X=X, y=y, alpha=alpha, **kwargs)

    return critical_for_linear_regression_from_values(
        coeffs=model["coef"].values,
        coeffs_se=model["se"].values,
        coeffs_names=model["names"].values,
        dof=model.df_resid_,
        confidence=confidence,
        alternative=alternative,
        variant=variant,
    )



from .corrtest import (
    critical_for_correlation_test,
    critical_for_correlation_test_from_values,
)
from .linreg import (
    critical_for_linear_regression,
    critical_for_linear_regression_from_values,
)
from .ttest import (
    critical_for_one_sample_ttest,
    critical_for_one_sample_ttest_from_values,
    critical_for_two_sample_ttest,
    critical_for_two_sample_ttest_from_values,
    determine_welch_correction,
)
from .utils import (
    get_alpha,
    get_bias_correction_factor_J,
)

__all__ = [
    "critical_for_correlation_test",
    "critical_for_correlation_test_from_values",
    "critical_for_linear_regression",
    "critical_for_linear_regression_from_values",
    "critical_for_one_sample_ttest",
    "critical_for_one_sample_ttest_from_values",
    "critical_for_two_sample_ttest",
    "critical_for_two_sample_ttest_from_values",
    "determine_welch_correction",
    "get_alpha",
    "get_bias_correction_factor_J",
]



import pytest

from critical_es_value import linreg


@pytest.mark.parametrize(
    "alternative, confidence, variant, expected, match",
    [
        ("two-sided", 0.95, "ttest", [2.1883620, 0.4416767, 0.5405673], None),
        ("two-sided", 0.99, "ttest", [2.9550478, 0.5964168, 0.7299533], None),
        ("two-sided", 0.95, "ztest", [2.0903826, 0.4219015, 0.5163645], None),
        ("two-sided", 0.99, "ztest", [2.7472284, 0.5544726, 0.6786179], None),
        ("less", 0.95, "ttest", [1.81662748, 0.366649610, 0.4487417218], None),
        ("greater", 0.95, "ttest", [1.81662748, 0.366649610, 0.4487417218], None),
        (
            "invalid",
            0.95,
            "ttest",
            None,
            r"alternative must be one of 'two-sided', 'greater', or 'less'",
        ),
        ("two-sided", 1.5, "ttest", None, r"confidence must be in \(0, 1\)"),
        (
            "two-sided",
            0.95,
            "invalid",
            None,
            r"variant must be one of 'ttest' or 'ztest'",
        ),
    ],
)
def test_critical_for_linear_regression(
    test_dataset4, alternative, confidence, variant, expected, match
):
    X = test_dataset4[["x1", "x2"]]
    y = test_dataset4["y"]

    if match:
        with pytest.raises(ValueError, match=match):
            linreg.critical_for_linear_regression(
                X=X,
                y=y,
                alternative=alternative,
                confidence=confidence,
                variant=variant,
            )
    else:
        result = linreg.critical_for_linear_regression(
            X=X,
            y=y,
            alternative=alternative,
            confidence=confidence,
            variant=variant,
        )
        for c_ex, c_act in zip(expected, result["coef_critical"]):
            assert pytest.approx(c_ex) == c_act



import numpy as np
from scipy import special as scipy_special


def get_alpha(confidence: float, alternative: str) -> float:
    r"""Calculate the significance level (alpha) corresponding to a given confidence level.

    .. math::

        \alpha = \begin{cases}
            1 - \text{conf} & \text{one-sided} \\
            \frac{1 - \text{conf}}{2} & \text{two-sided}
        \end{cases}

    Args:
        confidence (float): Confidence level between 0 and 1 (exclusive).
        alternative (str): The alternative hypothesis. Either "two-sided", "greater", or "less".

    Returns:
        float: The significance level (alpha).

    Raises:
        ValueError: If `confidence` is not in (0, 1) or if `alternative` is not one of "two-sided", "greater", or "less".

    Examples:
        >>> import critical_es_value as cev
        >>> cev.get_alpha(0.95, "two-sided")
        0.025
        >>> cev.get_alpha(0.95, "less")
        0.05
    """

    if confidence <= 0 or confidence >= 1:
        raise ValueError("confidence must be in (0, 1)")
    if alternative not in ("two-sided", "greater", "less"):
        raise ValueError("alternative must be one of 'two-sided', 'greater', or 'less'")

    alpha = 1 - confidence

    if alternative == "two-sided":
        return alpha / 2
    return alpha


def get_bias_correction_factor_J(dof: int) -> np.float64:
    r"""Calculate the bias correction factor J for Hedges' g.

    .. math::

        J(x) = \frac{\Gamma\left(\frac{x}{2}\right)}{\sqrt{\frac{x}{2}}\Gamma\left(\frac{x-1}{2}\right)}

    Args:
        dof (int): Degrees of freedom.

    Returns:
        np.float64: The bias correction factor J.

    Raises:
        ValueError: If dof is <= 1.

    Examples:
        >>> import critical_es_value as cev
        >>> cev.get_bias_correction_factor_J(10)
        0.92274560805
        >>> cev.get_bias_correction_factor_J(20)
        0.96194453374
    """
    if dof <= 1:
        raise ValueError("dof must be greater than 1.")

    num = scipy_special.loggamma(dof / 2)
    denom = np.log(np.sqrt(dof / 2)) + scipy_special.loggamma((dof - 1) / 2)
    return np.exp(num - denom)
